class PypactException(Exception):
    pass


class PypactFrozenException(PypactException):
    pass


class PypactInvalidOptionException(PypactException):
    pass


class PypactIncompatibleOptionException(PypactException):
    pass


class PypactOutOfRangeException(PypactException):
    pass


class PypactSerializeException(PypactException):
    pass


class PypactDeserializeException(PypactException):
    pass


class PypactFispactExecutableNotFoundException(PypactException):
    pass


class PypactTypeException(PypactException):
    pass


class PypactUnphysicalValueException(PypactException):
    pass


class PypactNotPrintLib4FileException(PypactException):
    pass


class PypactNotPrintLib5FileException(PypactException):
    pass


class PypactSpectrumDoesNotExistException(PypactException):
    pass



import math


class Edge:
    def __init__(self, lower, upper):
        assert upper >= lower
        self.lower = lower
        self.upper = upper

    @property
    def width(self):
        return (self.upper - self.lower)

    @property
    def log_ratio(self):
        if self.lower <= 0:
            return 1
        return math.log(self.upper/self.lower)


def get_overlap(edge1, edge2):
    IS_LOWER, IS_UPPER = True, False
    points = sorted([(edge1.lower, IS_LOWER), (edge1.upper, IS_UPPER),
                     (edge2.lower, IS_LOWER), (edge2.upper, IS_UPPER)], key=lambda x: x[0])
    # the second item must be a lower point and the third item must be an upper point
    # for overlap
    overlap = points[1][1] == IS_LOWER and points[2][1] == IS_UPPER
    overlap_edge = Edge(lower=points[1][0], upper=points[2][0])

    return overlap, overlap_edge


def _get_edges_from_bounds(bounds):
    return [Edge(bound, bounds[i+1])
            for i, bound in enumerate(bounds[:-1])]


def _convert_imp(input_bounds, input_values, output_bounds, cfunc):
    """
        Returns the output_values depending on the cfunc.

        output_bounds is a list of energies, units are irrelevant,
        as long as it matches the units of the input_bounds.

        Asserts both input and output bounds are of length greater than 1

        Assumes that input and output bounds are in ascending energy. If not
        then it will go unchecked and will produce odd results
    """

    assert len(output_bounds) > 1
    assert len(input_bounds) > 1

    input_edges = _get_edges_from_bounds(input_bounds)
    output_edges = _get_edges_from_bounds(output_bounds)

    def compute_overlap(oedge, last_overlap_index=0):
        output_value = 0.0
        prev_has_overlap = False
        for i, iedge in enumerate(input_edges):
            has_overlap, overlap_edge = get_overlap(iedge, oedge)
            if has_overlap:
                last_overlap_index = i
                output_value += cfunc(iedge,
                                      input_values[i],
                                      overlap_edge)

            if not has_overlap and prev_has_overlap:
                break

            prev_has_overlap = has_overlap

        return output_value, last_overlap_index

    last_index = 0
    output_values = []
    for edge in output_edges:
        output_value, last_index = compute_overlap(
            edge, last_overlap_index=last_index)
        output_values.append(output_value)

    return output_values


def by_energy(input_bounds, input_values, output_bounds):
    def cfunc(input_edge, input_value, overlapping_edge):
        return overlapping_edge.width*input_value/input_edge.width

    return _convert_imp(input_bounds, input_values, output_bounds, cfunc)


def by_lethargy(input_bounds, input_values, output_bounds):
    def cfunc(input_edge, input_value, overlapping_edge):
        return overlapping_edge.log_ratio*input_value/input_edge.log_ratio

    return _convert_imp(input_bounds, input_values, output_bounds, cfunc)



import json
import os

from ..util.exceptions import PypactSpectrumDoesNotExistException


_this_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)))
__SPECTRUM_JSON_LIB_FILE__ = os.path.join(
    _this_dir, 'data', 'spectrumlib.min.json')


class SpectrumLibJSONReader:

    def __init__(self, filename=__SPECTRUM_JSON_LIB_FILE__):
        self.filename = filename

    def __enter__(self):
        data = {}
        with open(self.filename, 'rt') as f:
            data = json.load(f)

        return data

    def __exit__(self, *args):
        pass


class SpectrumLibManager:
    def __init__(self, data):
        self._data = data

    def list(self):
        return list(self._data.keys())

    def get(self, name):
        if name not in self._data:
            raise PypactSpectrumDoesNotExistException(
                f"{name} does not exist in data")

        sdata = self._data[name]
        return sdata["energies"], sdata["values"]



import unittest

from tests.util.filetest import FileUnitTest
from tests.util.linestest import LinesUnitTest
from tests.util.numericaltest import NumericalUnitTest
from tests.util.propertyfindertest import PropertyFinderUnitTest

from tests.output.doseratetest import DoseRateUnitTest
from tests.output.gammaspectrumtest import GammaSpectrumUnitTest
from tests.output.rundatatest import RunDataUnitTest
from tests.output.nuclidestest import NuclidesUnitTest
from tests.output.timesteptest import TimeStepUnitTest
from tests.output.outputtest import OutputUnitTest
from tests.output.readertest import ReaderUnitTest

from tests.printlib.printlib4test import PrintLib4UnitTest
from tests.printlib.printlib5test import PrintLib5UnitTest

from tests.input.inputdatatest import InputDataUnitTest
from tests.input.filesfiletest import FilesFileUnitTest
from tests.input.fluxesfiletest import FluxesFileUnitTest
from tests.input.keywordstest import KeywordsUnitTest
from tests.input.groupconverttest import GroupConvertUnitTest
from tests.input.groupstructurestest import GroupStructuresUnitTest
from tests.input.projectilestest import ProjectilesUnitTest

from tests.library.nuclidelibtest import NuclideLibUnitTest
from tests.library.reactionlibtest import ReactionLibUnitTest


def main():
    unittest.TextTestRunner(verbosity=3).run(unittest.TestSuite())


if __name__ == '__main__':
    unittest.main()



import os
from setuptools import setup

VERSION = ""
with open(
    os.path.join(os.path.dirname(os.path.abspath(__file__)), ".VERSION"), "rt"
) as vfile:
    VERSION = vfile.readlines()[0].strip("\n").strip()


setup(
    name="pypact",
    version=VERSION,
    description="The package for reading and manipulating the fispact output text file.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/fispact/pypact",
    author="UKAEA",
    author_email="me@physicstom.com",
    license="Apache License 2.0",
    packages=[
        "pypact",
        "pypact.analysis",
        "pypact.library",
        "pypact.input",
        "pypact.output",
        "pypact.printlib",
        "pypact.tools",
        "pypact.util",
    ],
    install_requires=[],
    python_requires=">=3.6",
    scripts=["pypact/tools/fispactconverter.py"],
    setup_requires=[
        "pytest-runner",
    ],
    test_suite="tests.testsuite",
    tests_require=[
        "pytest",
        "pytest-cov>=2.3.1",
        "mock",
        "jsonschema",
        "numpy",
    ],
    package_data={"": ["library/data/*.json"]},
    include_package_data=True,
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "fispactconverter = pypact.tools.fispactconverter:main",
        ]
    },
)



import unittest
import math
import os

from pypact.util.numerical import are_values_the_same

REFERENCE_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             '..', 'reference')

class Tester(unittest.TestCase):

    def assertValueAndType(self, obj, objtype, parameter, parametertype, value, rel_tol=0.0):
        self.assertEqual(True, isinstance(obj, objtype))
        if rel_tol == 0.0:
            self.assertEqual(getattr(obj, parameter), value)
        else:
            self.assertIsClose(getattr(obj, parameter), value, rel_tol)
        self.assertEqual(True, isinstance(getattr(obj, parameter), parametertype))

    def assertIsClose(self, a, b, rel_tol):
        self.assertTrue(are_values_the_same(a, b, rel_tol=rel_tol, abs_tol=0.0))

    @staticmethod
    def _isnotfound(value):
        return value == 0.0
